/*
 * CPU-specific routines originally from U-Boot.
 * See: uboot-xburst/files/arch/mips/cpu/xburst/cpu.c
 * See: u-boot/arch/mips/include/asm/cacheops.h
 *
 * Copyright (C) 2000-2009 Wolfgang Denk, DENX Software Engineering, <wd@denx.de>
 * Copyright (C) 2015, 2016 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "cpu.h"
#include "cpu_op.h"
#include "memory.h"
#include "paging.h"
#include "sdram.h"

void flush_icache_all()
{
	u32 addr;

	flush_icache_tag();

	for (addr = KSEG0; addr < KSEG0 + CONFIG_SYS_ICACHE_SIZE; addr += CONFIG_SYS_CACHELINE_SIZE)
		flush_icache_region(addr);

	flush_icache_config();
}

void flush_dcache_all()
{
	u32 addr;

	for (addr = KSEG0; addr < KSEG0 + CONFIG_SYS_DCACHE_SIZE; addr += CONFIG_SYS_CACHELINE_SIZE)
		flush_dcache_region(addr);

	asm volatile ("sync");
}

void flush_cache_all()
{
	flush_dcache_all();
	flush_icache_all();
}

void init_registers(u32 *base, u32 got, void (*function)(), u32 args[], u8 nargs)
{
	u8 i;

	/* Provide arguments to the function. */

	for (i = 0; i < nargs; i++)
	{
		base[i+4] = args[i];
	}

	/* Store essential data for the function environment. */

	base[25] = (u32) function - 0x80000000;	/* store the function address as t9 */
	base[26] = got - 0x80000000;		/* store the global pointer */
	base[29] = (u32) function - 0x80000000;	/* store the function address as EPC (for the handler) */
}

void init_tlb(u8 first_random)
{
	u32 limit = configure_tlb(first_random), i;

	/* Reset the mappings. The total number is bits 30..25 of Config1. */

	for (i = 0; i < ((limit >> 25) & 0x3f); i++)
	{
		map_page_index(0, 0, 4096, 0, 0, i);
	}
}

void init_page_table(u32 page_table, u32 virtual, u32 physical, u32 pagesize, u8 flags, u8 asid)
{
	u32 lower = ((physical & 0xfffff000) >> 6) | flags;
	u32 upper = (((physical + pagesize) & 0xfffff000) >> 6) | flags;

	/*
	With a complete address space mapping involving pairs of 4KB pages
	described by two values for each entry, there would be...

	an address space of 0x100000000 requiring...

	0x100000000 / (8 * 1024) == 0x100000000 >> 13
	== 524288 entries
	== 0x80000 entries

	Thus, each task's entries would require...

	0x80000 * 8 == 0x400000 bytes

	The kseg2 region thus permits 256 tasks occupying 0x40000000 bytes.

	However, for more modest address spaces occupying as much as 32MB there
	would be...

	an address space of 0x02000000 requiring...

	0x02000000 / (8 * 1024) == 0x02000000 >> 13
	== 4096 entries
	== 0x1000 entries

	Thus, each task's entries would only require...

	0x1000 * 8 == 0x8000 bytes
	*/

	u32 base = page_table + STAGE2_PAGE_TABLE_TASK * asid;

	/* Each page table entry corresponds to a pair of 4KB pages and holds two values. */

	u32 entry = ((virtual & 0xffffe000) >> 13) * 8;
	u32 *address = (u32 *) (base + entry);

	/* The page tables should be permanently mapped to avoid hierarchical TLB miss handling. */

	*address = lower;
	*(address + 1) = upper;
}

void map_page(u32 virtual, u32 physical, u32 pagesize, u8 flags, u8 asid)
{
	u32 start = (virtual & 0xffffe000) | asid; /* VPN2 | ASID*/
	u32 lower = ((physical & 0xfffff000) >> 6) | flags;
	u32 upper = (((physical + pagesize) & 0xfffff000) >> 6) | flags;
	u32 pagemask = ((pagesize - 1) & 0xfffff000) << 1;

	map_page_op(lower, upper, start, pagemask);
}

void map_page_index(u32 virtual, u32 physical, u32 pagesize, u8 flags, u8 asid, u32 index)
{
	u32 start = (virtual & 0xffffe000) | asid; /* VPN2 | ASID*/
	u32 lower = ((physical & 0xfffff000) >> 6) | flags;
	u32 upper = (((physical + pagesize) & 0xfffff000) >> 6) | flags;
	u32 pagemask = ((pagesize - 1) & 0xfffff000) << 1;

	map_page_set_index(index);
	map_page_index_op(lower, upper, start, pagemask);
}
