/*
 * JzRISC LCD controller
 *
 * Copyright (C) Xiangfu Liu <xiangfu@sharism.cc>
 * Copyright (C) 2015, 2016 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "sdram.h"
#include "jzlcd.h"
#include "cpu.h"
#include "board.h"

#define align2(n) (n)=((((n)+1)>>1)<<1)
#define align4(n) (n)=((((n)+3)>>2)<<2)
#define align8(n) (n)=((((n)+7)>>3)<<3)

extern struct jzfb_info jzfb;
extern vidinfo_t panel_info;

static unsigned short lcd_get_panels()
{
	return ((jzfb.cfg & MODE_MASK) == MODE_STN_MONO_DUAL) ||
		((jzfb.cfg & MODE_MASK) == MODE_STN_COLOR_DUAL) ? 2 : 1;
}



/* Functions returning region sizes. */

static unsigned long lcd_get_size(vidinfo_t *vid)
{
	/* Lines must be aligned to a word boundary. */
	unsigned long line_length = ALIGN((vid->vl_col * NBITS(vid->vl_bpix)) / 8, sizeof(u32));
        return line_length * vid->vl_row;
}

static unsigned long lcd_get_aligned_size(vidinfo_t *vid)
{
	/* LCD_CTRL_BST_16 requires 16-word alignment. */
        return ALIGN(lcd_get_size(vid), 16 * sizeof(u32));
}

static unsigned long lcd_get_min_size(vidinfo_t *vid)
{
	/* Lines must be aligned to a word boundary. */
	unsigned long line_length = ALIGN((vid->vl_col * 32) / 8, sizeof(u32));
        return line_length * vid->vl_row;
}

static unsigned long lcd_get_aligned_min_size(vidinfo_t *vid)
{
	/* LCD_CTRL_BST_16 requires 16-word alignment. */
        return ALIGN(lcd_get_min_size(vid), 16 * sizeof(u32));
}

static unsigned long lcd_get_palette_size(vidinfo_t *vid)
{
	if (NBITS(vid->vl_bpix) < 12)
		return NCOLORS(vid->vl_bpix) * sizeof(u16);
	else
		return 0;
}

static unsigned long lcd_get_aligned_palette_size(vidinfo_t *vid)
{
	/* LCD_CTRL_BST_16 requires 16-word alignment. */
	return ALIGN(lcd_get_palette_size(vid), 16 * sizeof(u32));
}

static unsigned long lcd_get_descriptors_size()
{
	return 3 * sizeof(struct jz_fb_dma_descriptor);
}

static unsigned long lcd_get_total_size(vidinfo_t *vid)
{
	unsigned long size = lcd_get_aligned_size(vid) * lcd_get_panels();
	unsigned long min_size = lcd_get_aligned_min_size(vid);

        /* Round up to nearest full page, or MMU section if defined. */
	return ALIGN((size >= min_size ? size : min_size) + lcd_get_aligned_palette_size(vid) + lcd_get_descriptors_size(), PAGE_SIZE);
}



/* Functions returning addresses of each data region. */

static unsigned long lcd_get_palette(unsigned long addr)
{
        /* Allocate memory at the end of the region for the palette. */
	return addr - lcd_get_aligned_palette_size(&panel_info);
}

static unsigned long lcd_get_descriptors(unsigned long addr)
{
        /* Allocate memory before the palette for the descriptor array. */
	return lcd_get_palette(addr) - lcd_get_descriptors_size();
}

static unsigned long lcd_get_framebuffer(unsigned long addr, unsigned short panel)
{
        /* Allocate pages for the frame buffer and palette. */
        return addr - lcd_get_total_size(&panel_info) + (panel * lcd_get_aligned_size(&panel_info));
}



/* Initialisation functions. */

static void jz_lcd_desc_init(vidinfo_t *vid)
{
	struct jz_fb_dma_descriptor *descriptors;
	struct jz_fb_info * fbi;

	fbi = &vid->jz_fb;

	/* Allocate space for descriptors before the palette entries. */

	descriptors = (struct jz_fb_dma_descriptor *) lcd_get_descriptors(get_memory_size());
	fbi->dmadesc_fblow = (struct jz_fb_dma_descriptor *) &descriptors[0];
	fbi->dmadesc_fbhigh = (struct jz_fb_dma_descriptor *) &descriptors[1];
	fbi->dmadesc_palette = (struct jz_fb_dma_descriptor *) &descriptors[2];

	/* Populate descriptors. */

	if (lcd_get_panels() == 2)
	{
		fbi->dmadesc_fblow->fdadr = fbi->dmadesc_fblow;
		fbi->dmadesc_fblow->fsadr = lcd_get_framebuffer(get_memory_size(), 1);
		fbi->dmadesc_fblow->fidr  = 0;
		fbi->dmadesc_fblow->ldcmd = lcd_get_size(vid) / 4 ;

		fbi->fdadr1 = fbi->dmadesc_fblow; /* only used in dual-panel mode */
	}

	fbi->dmadesc_fbhigh->fsadr = fbi->screen;
	fbi->dmadesc_fbhigh->fidr = 0;
	fbi->dmadesc_fbhigh->ldcmd =  lcd_get_size(vid) / 4; /* length in words */

	if (NBITS(vid->vl_bpix) < 12)
	{
		fbi->dmadesc_palette->fsadr = fbi->palette;
		fbi->dmadesc_palette->fidr  = 0;
		fbi->dmadesc_palette->ldcmd = (lcd_get_palette_size(vid) / 4) | (1<<28);

		/* assume any mode with <12 bpp is palette driven */
		fbi->dmadesc_palette->fdadr = fbi->dmadesc_fbhigh;
		fbi->dmadesc_fbhigh->fdadr = fbi->dmadesc_palette;
		/* flips back and forth between pal and fbhigh */
		fbi->fdadr0 = fbi->dmadesc_palette;
	} else {
		/* palette shouldn't be loaded in true-color mode */
		fbi->dmadesc_fbhigh->fdadr = fbi->dmadesc_fbhigh;
		fbi->fdadr0 = fbi->dmadesc_fbhigh; /* no pal just fbhigh */
	}

	flush_cache_all();
}

static unsigned int jz_lcd_stn_init(unsigned int stnH)
{
	unsigned int val = 0;

	switch (jzfb.bpp) {
	case 1:
		/* val |= LCD_CTRL_PEDN; */
	case 2:
		val |= LCD_CTRL_FRC_2;
		break;
	case 4:
		val |= LCD_CTRL_FRC_4;
		break;
	case 8:
	default:
		val |= LCD_CTRL_FRC_16;
		break;
	}

	switch (jzfb.cfg & STN_DAT_PINMASK) {
	case STN_DAT_PIN1:
		/* Do not adjust the hori-param value. */
		break;
	case STN_DAT_PIN2:
		align2(jzfb.hsw);
		align2(jzfb.elw);
		align2(jzfb.blw);
		break;
	case STN_DAT_PIN4:
		align4(jzfb.hsw);
		align4(jzfb.elw);
		align4(jzfb.blw);
		break;
	case STN_DAT_PIN8:
		align8(jzfb.hsw);
		align8(jzfb.elw);
		align8(jzfb.blw);
		break;
	}

	REG_LCD_VSYNC = (0 << 16) | jzfb.vsw;
	REG_LCD_HSYNC = ((jzfb.blw+jzfb.w) << 16) | (jzfb.blw+jzfb.w+jzfb.hsw);

	/* Screen setting */
	REG_LCD_VAT = ((jzfb.blw + jzfb.w + jzfb.hsw + jzfb.elw) << 16) | (stnH + jzfb.vsw + jzfb.bfw + jzfb.efw);
	REG_LCD_DAH = (jzfb.blw << 16) | (jzfb.blw + jzfb.w);
	REG_LCD_DAV = (0 << 16) | (stnH);

	/* AC BIAs signal */
	REG_LCD_PS = (0 << 16) | (stnH+jzfb.vsw+jzfb.efw+jzfb.bfw);

	return val;
}

static void jz_lcd_tft_init()
{
	REG_LCD_VSYNC = (0 << 16) | jzfb.vsw;
	REG_LCD_HSYNC = (0 << 16) | jzfb.hsw;
	REG_LCD_DAV =((jzfb.vsw+jzfb.bfw) << 16) | (jzfb.vsw +jzfb.bfw+jzfb.h);
	REG_LCD_DAH = ((jzfb.hsw + jzfb.blw) << 16) | (jzfb.hsw + jzfb.blw + jzfb.w );
	REG_LCD_VAT = (((jzfb.blw + jzfb.w + jzfb.elw + jzfb.hsw)) << 16) \
		| (jzfb.vsw + jzfb.bfw + jzfb.h + jzfb.efw);
}

static void jz_lcd_samsung_init(unsigned int pclk)
{
	unsigned int total, tp_s, tp_e, ckv_s, ckv_e;
	unsigned int rev_s, rev_e, inv_s, inv_e;

	jz_lcd_tft_init();

	total = jzfb.blw + jzfb.w + jzfb.elw + jzfb.hsw;
	tp_s = jzfb.blw + jzfb.w + 1;
	tp_e = tp_s + 1;
	/* ckv_s = tp_s - jz_clocks.pixclk/(1000000000/4100); */
	ckv_s = tp_s - pclk/(1000000000/4100);
	ckv_e = tp_s + total;
	rev_s = tp_s - 11;	/* -11.5 clk */
	rev_e = rev_s + total;
	inv_s = tp_s;
	inv_e = inv_s + total;
	REG_LCD_CLS = (tp_s << 16) | tp_e;
	REG_LCD_PS = (ckv_s << 16) | ckv_e;
	REG_LCD_SPL = (rev_s << 16) | rev_e;
	REG_LCD_REV = (inv_s << 16) | inv_e;
	jzfb.cfg |= STFT_REVHI | STFT_SPLHI;
}

static void jz_lcd_sharp_init()
{
	unsigned int total, cls_s, cls_e, ps_s, ps_e;
	unsigned int spl_s, spl_e, rev_s, rev_e;

	jz_lcd_tft_init();

	total = jzfb.blw + jzfb.w + jzfb.elw + jzfb.hsw;
	spl_s = 1;
	spl_e = spl_s + 1;
	cls_s = 0;
	cls_e = total - 60;	/* > 4us (pclk = 80ns) */
	ps_s = cls_s;
	ps_e = cls_e;
	rev_s = total - 40;	/* > 3us (pclk = 80ns) */
	rev_e = rev_s + total;
	jzfb.cfg |= STFT_PSHI; 
	REG_LCD_SPL = (spl_s << 16) | spl_e;
	REG_LCD_CLS = (cls_s << 16) | cls_e;
	REG_LCD_PS = (ps_s << 16) | ps_e;
	REG_LCD_REV = (rev_s << 16) | rev_e;
}

static unsigned int jz_lcd_get_pixel_clock()
{
	unsigned int pclk;

	/* Derive pixel clock from frame clock. */

	if ( (jzfb.cfg & MODE_MASK) != MODE_8BIT_SERIAL_TFT) {
		pclk = jzfb.fclk * (jzfb.w + jzfb.hsw + jzfb.elw + jzfb.blw) *
			(jzfb.h + jzfb.vsw + jzfb.efw + jzfb.bfw);
	} else {
		/* serial mode: Hsync period = 3*Width_Pixel */
		pclk = jzfb.fclk * (jzfb.w*3 + jzfb.hsw + jzfb.elw + jzfb.blw) *
			(jzfb.h + jzfb.vsw + jzfb.efw + jzfb.bfw);
	}

	if (((jzfb.cfg & MODE_MASK) == MODE_STN_COLOR_SINGLE) ||
	    ((jzfb.cfg & MODE_MASK) == MODE_STN_COLOR_DUAL))
		pclk = (pclk * 3);

	if (((jzfb.cfg & MODE_MASK) == MODE_STN_COLOR_SINGLE) ||
	    ((jzfb.cfg & MODE_MASK) == MODE_STN_COLOR_DUAL) ||
	    ((jzfb.cfg & MODE_MASK) == MODE_STN_MONO_SINGLE) ||
	    ((jzfb.cfg & MODE_MASK) == MODE_STN_MONO_DUAL))
		pclk = pclk >> ((jzfb.cfg & STN_DAT_PINMASK) >> 4);

	if (((jzfb.cfg & MODE_MASK) == MODE_STN_COLOR_DUAL) ||
	    ((jzfb.cfg & MODE_MASK) == MODE_STN_MONO_DUAL))
		pclk >>= 1;

	return pclk;
}

static void jz_lcd_set_timing(unsigned int pclk)
{
	unsigned int val;

#ifdef CONFIG_CPU_JZ4730
        val = __cpm_get_pllout() / pclk;
        REG_CPM_CFCR2 = val - 1;
        val = pclk * 4 ;
        if ( val > 150000000 ) {
                val = 150000000;
        }
        val = __cpm_get_pllout() / val;
        val--;
        if ( val > 0xF )
                val = 0xF;
#else
	int pll_div;

	pll_div = ( REG_CPM_CPCCR & CPM_CPCCR_PCS ); /* clock source,0:pllout/2 1: pllout */
	pll_div = pll_div ? 1 : 2 ;
	val = ( __cpm_get_pllout()/pll_div ) / pclk;
	val--;
	if ( val > 0x1ff ) {
		val = 0x1ff;
	}
	__cpm_set_pixdiv(val);

	val = pclk * 3 ;	/* LCDClock > 2.5*Pixclock */
	if ( val > 150000000 ) {
		val = 150000000;
	}
	val = ( __cpm_get_pllout()/pll_div ) / val;
	val--;
	if ( val > 0x1f ) {
		val = 0x1f;
	}
#endif
	__cpm_set_ldiv( val );
	REG_CPM_CPCCR |= CPM_CPCCR_CE ; /* update divide */
}

static int jz_lcd_hw_init(vidinfo_t *vid)
{
	struct jz_fb_info *fbi = &vid->jz_fb;
	unsigned int val = 0;
	unsigned int pclk = jz_lcd_get_pixel_clock();

	/* Setting Control register */
	switch (jzfb.bpp) {
	case 1:
		val |= LCD_CTRL_BPP_1;
		break;
	case 2:
		val |= LCD_CTRL_BPP_2;
		break;
	case 4:
		val |= LCD_CTRL_BPP_4;
		break;
	case 8:
		val |= LCD_CTRL_BPP_8;
		break;
	case 15:
		val |= LCD_CTRL_RGB555;
	case 16:
		val |= LCD_CTRL_BPP_16;
		break;
#ifndef CONFIG_CPU_JZ4730
	case 17 ... 32:
		val |= LCD_CTRL_BPP_18_24;	/* target is 4bytes/pixel */
		break;
#endif
	default:
		/* printf("jz_lcd.c The BPP %d is not supported\n", jzfb.bpp); */
		val |= LCD_CTRL_BPP_16;
		break;
	}

	switch (jzfb.cfg & MODE_MASK) {
	case MODE_STN_MONO_DUAL:
	case MODE_STN_COLOR_DUAL:
		val |= jz_lcd_stn_init(jzfb.h >> 1);
		break;

	case MODE_STN_MONO_SINGLE:
	case MODE_STN_COLOR_SINGLE:
		val |= jz_lcd_stn_init(jzfb.h);
		break;

	case MODE_TFT_GEN:
	case MODE_TFT_CASIO:
	case MODE_8BIT_SERIAL_TFT:
	case MODE_TFT_18BIT:
		jz_lcd_tft_init();
		break;

	case MODE_TFT_SAMSUNG:
	{
		jz_lcd_samsung_init(pclk);
		break;
	}

	case MODE_TFT_SHARP:
	{
		jz_lcd_sharp_init();
		break;
	}

	default:
		break;
	}

	/* Configure the LCD panel */

	val |= LCD_CTRL_BST_16;		/* Burst Length is 16WORD=64Byte */
	val |= LCD_CTRL_OFUP;		/* OutFIFO underrun protect */
	REG_LCD_CTRL = val;
	REG_LCD_CFG = jzfb.cfg;

	/* Timing reset. */

	__cpm_stop_lcd();
	jz_lcd_set_timing(pclk);
	__cpm_start_lcd();
	udelay(1000);

	/* Configure DMA. */

	REG_LCD_DA0 = (unsigned long) fbi->fdadr0; /* frame descriptor */

	if (((jzfb.cfg & MODE_MASK) == MODE_STN_COLOR_DUAL) ||
	    ((jzfb.cfg & MODE_MASK) == MODE_STN_MONO_DUAL))
		REG_LCD_DA1 = (unsigned long) fbi->fdadr1; /* frame descriptor */

	return 0;
}

static inline u8 LCD_CODE(u8 bpp)
{       
        u8 code = 0;
        while (bpp && !(bpp & 1))
        {       
                bpp >>= 1;
                code += 1;
        }
        return code;
}

/* Public operations. */

void lcd_set_bpp(u8 bpp)
{
	jzfb.bpp = bpp;
	panel_info.vl_bpix = LCD_CODE(bpp);
}

void lcd_enable()
{
	/* Clear the disable bit and set the enable bit. */

	REG_LCD_CTRL &= ~(1<<4);	/* LCDCTRL.DIS */
	REG_LCD_CTRL |= 1<<3;		/* LCDCTRL.ENA*/
}

void lcd_disable()
{
	REG_LCD_CTRL |= (1<<4);		/* LCDCTRL.DIS, regular disable */
}

void lcd_quick_disable()
{
	REG_LCD_CTRL &= ~(1<<4);	/* LCDCTRL.ENA, quick disable */
}

static inline u16 rgb8_to_rgb16(u8 rgb)
{
	return ((((rgb & 0xe0) >> 5) * 4) << 11) | ((((rgb & 0x1c) >> 2) * 9) << 6) | ((rgb & 0x03) * 10);
}

static inline u16 rgb4_to_rgb16(u8 rgb)
{
	return ((((rgb & 8) >> 3) * 0x1f) << 11) | ((((rgb & 6) >> 1) * 0x15) << 5) | ((rgb & 1) * 0x1f);
}

static void lcd_init_palette(vidinfo_t *vid)
{
	u16 *palette = (u16 *) lcd_get_palette(get_memory_size());
	u16 *end = (u16 *) palette + NCOLORS(vid->vl_bpix);
	u8 value = 0;

	while (palette < end)
	{
		switch (vid->vl_bpix)
		{
			case LCD_COLOR4:
			*palette = rgb4_to_rgb16(value);
			break;

			case LCD_COLOR8:
			default:
			*palette = rgb8_to_rgb16(value);
			break;
		}

		value++;
		palette++;
	}
}

unsigned long lcd_ctrl_init()
{
	struct jz_fb_info *fbi = &panel_info.jz_fb;

	/* Start from the top of memory and obtain palette and framebuffer regions. */

	fbi->screen = lcd_get_framebuffer(get_memory_size(), 0);
	fbi->palette = lcd_get_palette(get_memory_size());

	if (NBITS(panel_info.vl_bpix) < 12)
		lcd_init_palette(&panel_info);

	jz_lcd_desc_init(&panel_info);
	jz_lcd_hw_init(&panel_info);

	return fbi->screen;
}
