/*
 * Task handling.
 *
 * Copyright (C) 2015, 2016 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "cpu.h"
#include "cpu_op.h"
#include "memory.h"
#include "mips.h"
#include "paging.h"
#include "tasks.h"

/* Task tables and data. */

static u32 stack_pointers[max_tasks];
static u32 registers[max_tasks][32];

u8 current_task;
u32 *current_stack_pointer;
u32 *current_registers;

/* A reference to locations for the symbol tables. */

extern u32 _got_start, _got_copy_start, _got_copy_end;

/* A reference to the start of the payload and end of memory locations. */

extern u32 _payload_start, _memory_end;

/* Task management functions. */

void init_tasks()
{
	current_task = 1;
	init_task();
}

void init_task()
{
	current_stack_pointer = &stack_pointers[current_task];
	current_registers = registers[current_task];
}

void start_task(unsigned short task, void (*function)(), u32 args[], u8 nargs)
{
	u32 virtual, physical, address;

	/*
	Each task employs a stack at a multiple of the given start address in
	physical memory, but at the same address in virtual memory. Task zero
	is never started.
	*/

	virtual = STAGE2_TASK_STACK;
	physical = STAGE2_TASK_STACK - STAGE2_TASK_STACK_SIZE * task;

	init_page_table(STAGE2_PAGE_TABLE, virtual - STAGE2_PAGESIZE * 2, physical - STAGE2_PAGESIZE * 2, STAGE2_PAGESIZE, TLB_WRITE, task);

	/*
	Subtract from the stack pointer to prevent the called function from
	reaching into unmapped memory.
	*/

	stack_pointers[task] = virtual - 12;

	/*
	Set the registers for the new task, initialising the global pointer and
	return address.
	*/

	init_registers(registers[task], (u32) &_got_copy_start, function, args, nargs);

	/* Map the global object table for the task. */

	init_page_table(STAGE2_PAGE_TABLE, user_address((u32) &_got_start), user_address((u32) &_got_copy_start), STAGE2_PAGESIZE, TLB_READ, task);

	/* Map all shared pages for the task. */

	for (address = (u32) &_payload_start; address < (u32) &_got_start; address += STAGE2_PAGESIZE * 2)
	{
		init_page_table(STAGE2_PAGE_TABLE, user_address(address), user_address(address), STAGE2_PAGESIZE, TLB_READ, task);
	}

	for (address = (u32) &_got_copy_end + STAGE2_PAGESIZE * 2; address < (u32) &_memory_end; address += STAGE2_PAGESIZE * 2)
	{
		init_page_table(STAGE2_PAGE_TABLE, user_address(address), user_address(address), STAGE2_PAGESIZE, TLB_WRITE, task);
	}
}

void start_task_now()
{
	invoke_task(current_task, current_registers, current_stack_pointer);
}

void switch_task()
{
	/* Switch the current task. */

	current_task++;
	if (current_task == max_tasks) current_task = 1;
	init_task();
}
